//
//  cbase64.cc
//					Author Tatsuo Baba
//
#include <stdio.h>
#include <string.h>
#include "cbase64.h"

int DecodeValue(const int c);

const char szBASE64CODE[] =
  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
#define ENCODE_BYTE(b) szBASE64CODE[((uchar)b) & 077];

// Calculate decoded character value.  Note that only legal
// characters should be given as input.
int DecodeValue(const int c)
{
	// Note that this works only on ASCII machines.
	if ('A' <= c && c <= 'Z')
		return c - 'A';
	if ('a' <= c && c <= 'z')
		return c - 'a' + 26;
	if ('0' <= c && c <= '9')
		return c - '0' + 52;
	if (c == '+')
		return 62;
	if (c == '/')
		return 63;
	if (c == '=')
		return -1;
	return -2;
}

/*********************************************************************
	Class: CBase64Enc
*********************************************************************/

/*
	Function: CBase64Enc::OpenFile
	Description:
		Opens the source file from which the data will be drawn.
*/
int CBase64Enc::OpenFile(char * pszFileName)
{
	// Open the file.  Handle any errors.
	m_hfile = fopen(pszFileName, "rb");
	if (m_hfile == NULL) {
		return RESULT_ERROR;
	}
	else
		return RESULT_OK;
}

/*
	Function: CBase64Enc::EncodeLine
	Description:
		Reads a line's-worth of data from the file, encodes it,
		and returns it to the caller.
*/
int CBase64Enc::EncodeLine(char * pszLine,int size)
{
	unsigned char buff[256];  // must be unsigned char for ENCODE_BYTE

//	ASSERT(size < sizeof(buff));

	memset(buff,0,sizeof(buff));			// clear buffer
	// Read a line's-worth of bytes from the file.
	int nBytes = fread(buff,1,size,m_hfile);
	int iByteIndex = 0;
	int iCharIndex = 0;
	while (nBytes > 0 && iByteIndex < nBytes) {
		pszLine[iCharIndex] =
			ENCODE_BYTE(buff[iByteIndex] >> 2);
		pszLine[iCharIndex + 1] =
			ENCODE_BYTE((buff[iByteIndex] << 4) |
				(buff[iByteIndex + 1] >> 4));
		if (iByteIndex + 1 < nBytes) {
			pszLine[iCharIndex + 2] =
				ENCODE_BYTE((buff[iByteIndex + 1] << 2) |
					(buff[iByteIndex + 2] >> 6));
		}
		else {
			pszLine[iCharIndex + 2] = '=';
		}
		if (iByteIndex + 2 < nBytes) {
			pszLine[iCharIndex + 3] =
				ENCODE_BYTE(buff[iByteIndex + 2]);
		}
		else {
			pszLine[iCharIndex + 3] = '=';
		}
		iByteIndex += 3;
		iCharIndex += 4;
    }

	// Add the line terminator.
	pszLine[iCharIndex] = '\0';
	// See if we hit the end of the file.
	if (nBytes == 0) {
		// Yup, so close it up, handling any error as appropriate.
		if (fclose(m_hfile)) {
			m_hfile = NULL;
			return RESULT_ERROR;
		}
		else {
			m_hfile = NULL;
			return RESULT_DONE;
		}
	}
	else {
		return RESULT_OK;
	}
}



/*
	Function: CBase64Enc::EncodeLine
	Description:
		passed data encodes it,
		and returns it to the caller.
*/
int CBase64Enc::EncodeLine(int nDataLen,char* pszData,char * pszLine, int nLen)
{

	// Make sure the line buffer passed to us will be able to hold
	// all the encoded data.  If not, error.
	if (nLen < ((nDataLen / 3) * 4 + 1))
		return RESULT_SMALLBUFFER;

	unsigned char *tp = (unsigned char*)pszData;	// must be unsigned char for ENCODE_BYTE

	// Zero out the byte buffer so that when we reach the final
	// line and it's not full that it'll be padded with zero.
	memset(pszLine,0,nLen);
	// Copy a line's-worth of bytes.
	int nBytes = nDataLen;
	int iByteIndex = 0;
	int iCharIndex = 0;
	while (nBytes > 0 && iByteIndex < nBytes) {
		pszLine[iCharIndex] =
			ENCODE_BYTE(tp[iByteIndex] >> 2);
		pszLine[iCharIndex + 1] =
			ENCODE_BYTE((tp[iByteIndex] << 4) |
				(tp[iByteIndex + 1] >> 4));
		if (iByteIndex + 1 < nBytes) {
			pszLine[iCharIndex + 2] =
				ENCODE_BYTE((tp[iByteIndex + 1] << 2) |
					(tp[iByteIndex + 2] >> 6));
		}
		else {
			pszLine[iCharIndex + 2] = '=';
		}
		if (iByteIndex + 2 < nBytes) {
			pszLine[iCharIndex + 3] =
				ENCODE_BYTE(tp[iByteIndex + 2]);
		}
		else {
			pszLine[iCharIndex + 3] = '=';
		}
		iByteIndex += 3;
		iCharIndex += 4;
    }

	// Add the line terminator.
	pszLine[iCharIndex] = '\0';
	// See if we hit the end of the file.
	if (nBytes == 0) 
			return RESULT_DONE;
	else 
		return RESULT_OK;
	
}




/*********************************************************************
	Class: CBase64Dec
*********************************************************************/

/*
	Function: CBase64Dec::OpenFile
	Description:
		Open a file that we can write the characters to.
*/
int CBase64Dec::OpenFile(char * pszFileName)
{
	// Open the file.  Handle any errors.
	m_hfile = fopen(pszFileName, "wb");
	if (m_hfile == NULL) {
		return RESULT_ERROR;
	}
	else
		return RESULT_OK;
}

/*
	Function: CBase64Dec::DecodeLine
	Description:
		Decode a line of data and write the bytes to the open file.
*/
int CBase64Dec::DecodeLine(char * pszLine, int nLineLen)
{
	uchar * uchBuffer = new uchar[nLineLen];
	int nDecodedLen = DecodeLine(pszLine, nLineLen,
		uchBuffer, nLineLen);
	if (nDecodedLen >= 0) {
		if (fwrite(uchBuffer,1, nDecodedLen,m_hfile) == 0)
			nDecodedLen = RESULT_ERROR;
	}
	delete[] uchBuffer;
    return nDecodedLen;
}

/*
	Function: CBase64Dec::DecodeLine
	Description:
		Simply decode a line of bytes.  Note that a file doesn't
		have to be opened to use this member function. Returns
        the count of the decoded bytes or -1 on error.
*/
int CBase64Dec::DecodeLine(char * pszLine, int nLineLen,
	uchar * uchBuffer, int nBufferLen)
{
	int iLineIndex = 0;
	int iBufferIndex = 0;

	while (iLineIndex < nLineLen) {
		// Group together four characters for decode.
		while (iLineIndex < nLineLen && m_iChars < 4) {
			int c = pszLine[iLineIndex++];
			// Ignore characters that aren't BASE64 characters
            // (e.g., spaces, CRLF, etc.).
			if (DecodeValue(c) != -2)
				m_uchStoredChars[m_iChars++] = (uchar) c;
		}

		if (m_iChars == 4) {
			// We've got four characters, so decode them.
			m_iChars = 0;

			// Decode first byte.
			if (iBufferIndex == nBufferLen) return RESULT_ERROR;
			uchBuffer[iBufferIndex++] = (uchar)
				(((uchar)DecodeValue(m_uchStoredChars[0]) << 2) |
				((uchar)DecodeValue(m_uchStoredChars[1]) >> 4));

			// Decode second byte.
			if (iBufferIndex == nBufferLen) return RESULT_ERROR;
			if (m_uchStoredChars[2] == '=') goto retcv64;
			uchBuffer[iBufferIndex++] = (uchar)
				(((uchar)DecodeValue(m_uchStoredChars[1]) << 4) |
				((uchar)DecodeValue(m_uchStoredChars[2]) >> 2));

			// Decode third byte.
			if (iBufferIndex == nBufferLen) return RESULT_ERROR;
			if (m_uchStoredChars[3] == '=') goto retcv64;
			uchBuffer[iBufferIndex++] = (uchar)
				(((uchar)DecodeValue(m_uchStoredChars[2]) << 6) |
				((uchar)DecodeValue(m_uchStoredChars[3])));
		}
	}
retcv64:
    // Return the count of decoded bytes.
	uchBuffer[iBufferIndex] = 0;
    return iBufferIndex;
}
